"""
Module to provide tests related to the MD044 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest

# pylint: disable=too-many-lines


@pytest.mark.rules
def test_md044_bad_configuration_names():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=$#1",
        "--strict-config",
        "scan",
        "test/resources/rules/md044/good_paragraph_text.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md044.names' must be of type 'str'."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_configuration_names_empty_elements():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=,,",
        "--strict-config",
        "scan",
        "test/resources/rules/md044/good_paragraph_text.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "Elements in the comma-separated list cannot be empty."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_configuration_names_repeated_elements():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=one,two,One",
        "--strict-config",
        "scan",
        "test/resources/rules/md044/good_paragraph_text.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "Element `One` is already present in the list as `one`."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_configuration_code_blocks():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.code_blocks=one",
        "--strict-config",
        "scan",
        "test/resources/rules/md044/good_paragraph_text.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md044.code_blocks' must be of type 'bool'."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_paragraph_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=paragraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_paragraph_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_paragraph_text.md:1:11: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_paragraph_text_prefix():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_prefix.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_paragraph_text_suffix():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_suffix.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_paragraph_text_start():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_start.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_paragraph_text_start.md:1:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_paragraph_text_end():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_end.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_paragraph_text_end.md:1:20: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_paragraph_text_followed_non_alpha():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_followed_non_alpha.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_paragraph_text_followed_non_alpha.md:1:11: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bar_paragraph_text_multiples():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_multiples.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_paragraph_text_multiples.md:1:11: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_paragraph_text_multiples.md:1:26: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bar_paragraph_text_multiples_on_multiple_lines():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_paragraph_text_multiples_on_multiple_lines.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_paragraph_text_multiples_on_multiple_lines.md:1:33: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_paragraph_text_multiples_on_multiple_lines.md:3:39: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bar_atx_heading_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_atx_heading_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_atx_heading_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_atx_heading_text.md:3:8: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_setext_heading_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_setext_heading_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_setext_heading_text.md:1:25: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_setext_heading_text.md:4:8: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_indented_code_block_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_indented_code_block_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_indented_code_block_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_indented_code_block_text.md:3:12: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_bad_fenced_code_block_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_fenced_code_block_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_fenced_code_block_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_fenced_code_block_text.md:4:8: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_indented_code_block_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "--set",
        "plugins.md044.code_blocks=$!False",
        "--strict-config",
        "scan",
        "test/resources/rules/md044/good_indented_code_block_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_indented_code_block_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_fenced_code_block_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "--set",
        "plugins.md044.code_blocks=$!False",
        "--strict-config",
        "scan",
        "test/resources/rules/md044/good_fenced_code_block_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_fenced_code_block_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_html_block_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_html_block_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_html_block_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_html_block_text.md:4:8: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_block_quote_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_block_quote_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_block_quote_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_block_quote_text.md:3:10: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_code_span_text():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_code_span_text.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_code_span_text.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_code_span_text.md:3:7: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_code_span_text_multiple_lines():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_code_span_text_multiple_lines.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_code_span_text_multiple_lines.md:1:27: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_code_span_text_multiple_lines.md:3:7: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_code_span_text_multiple_lines.md:4:6: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_link():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_link.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_link.md:1:4: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_link.md:1:41: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_link_multiple_lines_x():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_link_multiple_lines.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_link_multiple_lines.md:1:4: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_link_multiple_lines.md:3:2: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_link_multiple_lines_two():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_link_multiple_lines_two.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_link_multiple_lines_two.md:1:4: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_link_multiple_lines_two.md:4:13: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_link_multiple_lines_three():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_link_multiple_lines_three.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_link_multiple_lines_three.md:2:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_link_multiple_lines_three.md:4:2: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_image():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_image.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_image.md:1:5: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_image.md:1:42: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_image_multiple_lines():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_image_multiple_lines.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_image_multiple_lines.md:1:5: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_image_multiple_lines.md:3:2: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_inline_image_multiple_lines_two():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_inline_image_multiple_lines_two.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_inline_image_multiple_lines_two.md:1:5: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_inline_image_multiple_lines_two.md:4:13: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_full_link():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_full_link.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_full_link.md:2:4: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_full_link.md:5:16: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_full_link_multiple():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_full_link_multiple.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_full_link_multiple.md:3:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_full_link_multiple.md:9:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_full_image():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_full_image.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_full_image.md:2:5: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_full_image.md:5:16: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_full_image_multiple():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_full_image_multiple.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_full_image_multiple.md:3:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_full_image_multiple.md:9:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_collapsed_link():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_collapsed_link.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_collapsed_link.md:3:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_link.md:7:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_link.md:7:21: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_collapsed_link_multiple():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_collapsed_link_multiple.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_collapsed_link_multiple.md:3:6: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_link_multiple.md:7:6: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_link_multiple.md:8:5: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_collapsed_image():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_collapsed_image.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_collapsed_image.md:3:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_image.md:7:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_image.md:7:21: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md044_good_collapsed_image_multiple():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md044.names=ParaGraph",
        "scan",
        "test/resources/rules/md044/good_collapsed_image_multiple.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md044/good_collapsed_image_multiple.md:3:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_image_multiple.md:7:1: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)\n"
        + "test/resources/rules/md044/good_collapsed_image_multiple.md:8:5: "
        + "MD044: Proper names should have the correct capitalization "
        + "[Expected: ParaGraph; Actual: paragraph] (proper-names)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_360():
    """
    Test case 360:  Rule 1:
    """

    # Arrange
    source_markdown = """*foo bar*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo bar:]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_361():
    """
    Test case 361:  This is not emphasis, because the opening * is followed by whitespace, and hence not part of a left-flanking delimiter run:
    """

    # Arrange
    source_markdown = """a * foo bar*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a :]",
        "[text(1,3):*:]",
        "[text(1,4): foo bar:]",
        "[text(1,12):*:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a * foo bar*</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_362():
    """
    Test case 362:  This is not emphasis, because the opening * is preceded by an alphanumeric and followed by punctuation, and hence not part of a left-flanking delimiter run:
    """

    # Arrange
    source_markdown = """a*"foo"*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[text(1,2):*:]",
        '[text(1,3):\a"\a&quot;\afoo\a"\a&quot;\a:]',
        "[text(1,8):*:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a*&quot;foo&quot;*</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_363():
    """
    Test case 363:  Unicode nonbreaking spaces count as whitespace, too:
    """

    # Arrange
    source_markdown = """*\u00A0a\u00A0*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[text(1,2):\u00A0a\u00A0:]",
        "[text(1,5):*:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*\u00A0a\u00A0*</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_364():
    """
    Test case 364:  (part 1) Intraword emphasis with * is permitted:
    """

    # Arrange
    source_markdown = """foo*bar*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo:]",
        "[emphasis(1,4):1:*]",
        "[text(1,5):bar:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo<em>bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_365():
    """
    Test case 365:  (part 2) Intraword emphasis with * is permitted:
    """

    # Arrange
    source_markdown = """5*6*78"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):5:]",
        "[emphasis(1,2):1:*]",
        "[text(1,3):6:]",
        "[end-emphasis(1,4)::]",
        "[text(1,5):78:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>5<em>6</em>78</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide classes to deal with plugins.
"""
import argparse
import inspect
import logging
import os
import re
import sys
from abc import ABC, abstractmethod

from application_properties import ApplicationPropertiesFacade
from columnar import columnar

from pymarkdown.extensions.pragma_token import PragmaExtension
from pymarkdown.parser_helper import ParserHelper

LOGGER = logging.getLogger(__name__)

# pylint: disable=too-many-lines


# pylint: disable=too-few-public-methods
class ScanContext:
    """
    Class to provide context when reporting any errors.
    """

    def __init__(self, owning_manager, scan_file):
        self.owning_manager, self.scan_file, self.line_number = (
            owning_manager,
            scan_file,
            0,
        )
        self.__reported = []

    # pylint: disable=too-many-arguments
    def add_triggered_rule(
        self,
        scan_file,
        line_number,
        column_number,
        rule_id,
        rule_name,
        rule_description,
        extra_error_information,
    ):
        """
        Add the triggering information for a rule.
        """
        new_entry = (
            scan_file,
            line_number,
            column_number,
            rule_id,
            rule_name,
            rule_description,
            extra_error_information,
        )
        self.__reported.append(new_entry)

    # pylint: enable=too-many-arguments

    def report_on_triggered_rules(self):
        """
        Report on the various points where rules were triggered,
        in sorted order.
        """
        reported_and_sorted = sorted(self.__reported)

        for next_entry in reported_and_sorted:
            self.owning_manager.log_scan_failure(
                scan_file=next_entry[0],
                line_number=next_entry[1],
                column_number=next_entry[2],
                rule_id=next_entry[3],
                rule_name=next_entry[4],
                rule_description=next_entry[5],
                extra_error_information=next_entry[6],
            )
        self.__reported.clear()


# pylint: enable=too-few-public-methods


class BadPluginError(Exception):
    """
    Class to allow for a critical error within a plugin to be encapsulated
    and reported.
    """

    # pylint: disable=too-many-arguments, too-many-branches
    def __init__(
        self,
        plugin_id=None,
        plugin_action=None,
        file_name=None,
        class_name=None,
        field_name=None,
        is_constructor=False,
        is_empty=False,
        formatted_message=None,
        line_number=0,
        column_number=0,
        actual_line=None,
        actual_token=None,
        cause=None,
    ):

        if not formatted_message:
            if file_name:
                if class_name:
                    if is_constructor:
                        formatted_message = f"Plugin file named '{file_name}' threw an exception in the constructor for the class '{class_name}'."
                    else:
                        formatted_message = f"Plugin file named '{file_name}' does not contain a class named '{class_name}'."
                else:
                    formatted_message = (
                        f"Plugin file named '{file_name}' cannot be loaded."
                    )
            elif class_name:
                if field_name:
                    if is_empty:
                        formatted_message = f"Plugin class '{class_name}' returned an empty value for field name '{field_name}'."
                    else:
                        formatted_message = f"Plugin class '{class_name}' returned an improperly typed value for field name '{field_name}'."
                else:
                    formatted_message = f"Plugin class '{class_name}' had a critical failure loading the plugin details."
            else:
                if cause and isinstance(cause, ValueError):
                    formatted_message = str(cause)
                else:
                    formatted_message = f"Plugin id '{plugin_id.upper()}' had a critical failure during the '{str(plugin_action)}' action."
            if line_number:
                position_message = (
                    f"({line_number},{column_number})"
                    if column_number
                    else f"(Line {line_number})"
                )
                formatted_message = f"{position_message}: {formatted_message}"
            if actual_line:
                formatted_message = f"{formatted_message}\nActual Line: {actual_line}"
            if actual_token:
                formatted_message = f"{formatted_message}\nActual Token: {ParserHelper.make_value_visible(actual_token)}"
        super().__init__(formatted_message)

    # pylint: enable=too-many-arguments, too-many-branches


class Plugin(ABC):
    """
    Class to provide structure to scan through a file.
    Based off of concepts from https://github.com/hiddenillusion/example-code/commit/3e2daada652fe9b487574c784e0924bd5fcfe667
    """

    def __init__(self):
        (
            self.__plugin_specific_facade,
            self.__is_next_token_implemented_in_plugin,
            self.__is_next_line_implemented_in_plugin,
            self.__is_starting_new_file_implemented_in_plugin,
            self.__is_completed_file_implemented_in_plugin,
        ) = (None, True, True, True, True)

    @abstractmethod
    def get_details(self):
        """
        Get the details for the plugin.
        """

    @property
    def plugin_configuration(self):
        """
        Get the configuration facade associated with this plugin.
        """
        return self.__plugin_specific_facade

    def set_configuration_map(self, plugin_specific_facade):
        """
        Set the configuration map with values for the plugin.
        """
        self.__plugin_specific_facade = plugin_specific_facade

        self.__is_next_token_implemented_in_plugin = (
            "next_token" in self.__class__.__dict__.keys()
        )
        self.__is_next_line_implemented_in_plugin = (
            "next_line" in self.__class__.__dict__.keys()
        )
        self.__is_starting_new_file_implemented_in_plugin = (
            "starting_new_file" in self.__class__.__dict__.keys()
        )
        self.__is_completed_file_implemented_in_plugin = (
            "completed_file" in self.__class__.__dict__.keys()
        )

    @property
    def is_starting_new_file_implemented_in_plugin(self):
        """
        Return whether the starting_new_file function is implemented in the plugin.
        """
        return self.__is_starting_new_file_implemented_in_plugin

    @property
    def is_next_line_implemented_in_plugin(self):
        """
        Return whether the next_line function is implemented in the plugin.
        """
        return self.__is_next_line_implemented_in_plugin

    @property
    def is_next_token_implemented_in_plugin(self):
        """
        Return whether the next_token function is implemented in the plugin.
        """
        return self.__is_next_token_implemented_in_plugin

    @property
    def is_completed_file_implemented_in_plugin(self):
        """
        Return whether the completed_file function is implemented in the plugin.
        """
        return self.__is_completed_file_implemented_in_plugin

    def report_next_line_error(
        self, context, column_number, line_number_delta=0, extra_error_information=None
    ):
        """
        Report an error with the current line being processed.
        """
        context.add_triggered_rule(
            context.scan_file,
            context.line_number + line_number_delta,
            column_number,
            self.get_details().plugin_id,
            self.get_details().plugin_name,
            self.get_details().plugin_description,
            extra_error_information,
        )

    # pylint: disable=too-many-arguments
    def report_next_token_error(
        self,
        context,
        token,
        extra_error_information=None,
        line_number_delta=0,
        column_number_delta=0,
        use_original_position=False,
    ):
        """
        Report an error with the current token being processed.
        """
        context.add_triggered_rule(
            context.scan_file,
            (token.original_line_number if use_original_position else token.line_number)
            + line_number_delta,
            (
                token.original_column_number
                if use_original_position
                else token.column_number
            )
            + column_number_delta
            if column_number_delta >= 0
            else -column_number_delta,
            self.get_details().plugin_id,
            self.get_details().plugin_name,
            self.get_details().plugin_description,
            extra_error_information,
        )

    # pylint: enable=too-many-arguments

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """

    def completed_file(self, context):
        """
        Event that the file being currently scanned is now completed.
        """

    def next_line(self, context, line):
        """
        Event that a new line is being processed.
        """

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class PluginDetails:
    """
    Class to provide details about a plugin, supplied by the plugin.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        plugin_id,
        plugin_name,
        plugin_description,
        plugin_enabled_by_default,
        plugin_version,
        plugin_interface_version,
        plugin_url=None,
        plugin_configuration=None,
    ):
        (
            self.plugin_id,
            self.plugin_name,
            self.plugin_description,
            self.plugin_enabled_by_default,
            self.plugin_version,
            self.plugin_interface_version,
            self.plugin_url,
            self.plugin_configuration,
        ) = (
            plugin_id,
            plugin_name,
            plugin_description,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            plugin_url,
            plugin_configuration,
        )

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods,too-many-instance-attributes


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class FoundPlugin:
    """
    Encapsulation of a plugin that was discovered.  While similar to the PluginDetails
    class, this is meant for an internal representation of the plugin, and not the
    external information provided.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        plugin_id,
        plugin_name,
        plugin_description,
        plugin_instance,
        plugin_enabled_by_default,
        plugin_version,
        plugin_interface_version,
        instance_file_name,
        plugin_url,
        plugin_configuration,
    ):
        """
        Initializes a new instance of the FoundPlugin class.
        """
        (
            self.__plugin_id,
            self.__plugin_names,
            self.__plugin_description,
            self.__plugin_instance,
            self.__plugin_enabled_by_default,
            self.__plugin_version,
            self.__plugin_interface_version,
            self.__plugin_file_name,
            self.__plugin_url,
            self.__plugin_configuration,
        ) = (
            plugin_id.strip().lower(),
            [],
            plugin_description,
            plugin_instance,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            instance_file_name,
            plugin_url,
            plugin_configuration,
        )
        for next_name in plugin_name.lower().split(","):
            next_name = next_name.strip()
            if next_name:
                self.__plugin_names.append(next_name)

    # pylint: enable=too-many-arguments

    @property
    def plugin_id(self):
        """
        Gets the id associated with the plugin.
        """
        return self.__plugin_id

    @property
    def plugin_names(self):
        """
        Gets the names associated with the plugin.
        """
        return self.__plugin_names

    @property
    def plugin_identifiers(self):
        """
        Gets the identifiers (id+names) for the plugin.
        """
        plugin_keys = [self.plugin_id]
        plugin_keys.extend(self.plugin_names)
        return plugin_keys

    @property
    def plugin_description(self):
        """
        Gets the description of the plugin.
        """
        return self.__plugin_description

    @property
    def plugin_instance(self):
        """
        Gets the actual instance of the plugin.
        """
        return self.__plugin_instance

    @property
    def plugin_file_name(self):
        """
        Gets the filename where the plugin's class is stored.
        """
        return self.__plugin_file_name

    @property
    def plugin_version(self):
        """
        Gets the version of the plugin.
        """
        return self.__plugin_version

    @property
    def plugin_url(self):
        """
        Gets the optional url for the plugin.
        """
        return self.__plugin_url

    @property
    def plugin_configuration(self):
        """
        Gets the optional configuration items for the plugin.
        """
        return self.__plugin_configuration

    @property
    def plugin_enabled_by_default(self):
        """
        Gets a value indicating whether the plugin is enabled by default.
        """
        return self.__plugin_enabled_by_default


# pylint: enable=too-few-public-methods,too-many-instance-attributes


# pylint: disable=too-many-instance-attributes
class PluginManager:
    """
    Manager object to take care of load and accessing plugin modules.
    """

    __plugin_prefix = "plugins"
    __root_subparser_name = "pm_subcommand"
    __argparse_subparser = None
    __id_regex = re.compile("^[a-z]{2,3}[0-9]{3,3}$")
    __name_regex = re.compile("^[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9]$")
    __filter_regex = re.compile("^[a-zA-Z0-9-]+$")
    __version_regex = re.compile("^(0|[1-9]\\d*)\\.(0|[1-9]\\d*)\\.(0|[1-9]\\d*)$")

    def __init__(self):
        (
            self.__registered_plugins,
            self.__enabled_plugins,
            self.__enabled_plugins_for_starting_new_file,
            self.__enabled_plugins_for_next_token,
            self.__enabled_plugins_for_next_line,
            self.__enabled_plugins_for_completed_file,
            self.__loaded_classes,
            self.number_of_scan_failures,
            self.number_of_pragma_failures,
            self.__show_stack_trace,
            self.__document_pragmas,
            self.__all_ids,
        ) = (None, None, None, None, None, None, None, None, None, False, None, None)

    # pylint: disable=too-many-arguments
    def initialize(
        self,
        directory_to_search,
        additional_paths,
        enable_rules_from_command_line,
        disable_rules_from_command_line,
        properties,
        show_stack_trace,
    ):
        """
        Initializes the manager by scanning for plugins, loading them, and registering them.
        """
        (
            self.number_of_scan_failures,
            self.number_of_pragma_failures,
            self.__loaded_classes,
            self.__show_stack_trace,
        ) = (0, 0, [], show_stack_trace)

        plugin_files = self.__find_eligible_plugins_in_directory(directory_to_search)
        self.__load_plugins(directory_to_search, plugin_files)

        all_additional_paths = []
        if additional_paths:
            all_additional_paths.extend(additional_paths)
        more_paths = properties.get_string_property("plugins.additional_paths")
        if more_paths:
            all_additional_paths.extend(more_paths.split(","))

        if all_additional_paths:
            for next_additional_plugin in all_additional_paths:
                if not os.path.exists(next_additional_plugin):
                    formatted_message = (
                        f"Plugin path '{next_additional_plugin}' does not exist."
                    )
                    raise BadPluginError(formatted_message=formatted_message)
                if os.path.isdir(next_additional_plugin):
                    plugin_files = self.__find_eligible_plugins_in_directory(
                        next_additional_plugin
                    )
                    self.__load_plugins(next_additional_plugin, plugin_files)
                else:
                    self.__load_plugins(
                        os.path.dirname(next_additional_plugin),
                        [os.path.basename(next_additional_plugin)],
                    )

        self.__register_plugins(
            enable_rules_from_command_line, disable_rules_from_command_line, properties
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def argparse_subparser_name():
        """
        Gets the name of the subparser used to handle these plugins.
        """
        return "plugins"

    @staticmethod
    def __list_filter_type(argument):
        test_argument = argument.replace("*", "").replace("?", "")
        if PluginManager.__filter_regex.match(test_argument):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid pattern for an id or a name."
        )

    @staticmethod
    def __info_filter_type(argument):
        if PluginManager.__id_regex.match(argument) or PluginManager.__name_regex.match(
            argument
        ):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid id or name."
        )

    @staticmethod
    def add_argparse_subparser(subparsers):
        """
        Populate the argparse tree to allow for plugin support.
        """

        new_sub_parser = subparsers.add_parser(
            PluginManager.argparse_subparser_name(), help="plugin commands"
        )
        PluginManager.__argparse_subparser = new_sub_parser
        plugin_subparsers = new_sub_parser.add_subparsers(
            dest=PluginManager.__root_subparser_name
        )

        sub_sub_parser = plugin_subparsers.add_parser(
            "list", help="list the available plugins"
        )
        sub_sub_parser.add_argument(
            "--all",
            dest="show_all",
            action="store_true",
            default=False,
            help="show all loaded plugins (default is False)",
        )
        sub_sub_parser.add_argument(
            dest="list_filter",
            default=None,
            help="filter",
            nargs="?",
            type=PluginManager.__list_filter_type,
        )
        sub_sub_parser = plugin_subparsers.add_parser(
            "info", help="information on a specific plugin"
        )
        sub_sub_parser.add_argument(
            dest="info_filter",
            default=None,
            type=PluginManager.__info_filter_type,
            help="an id",
        )

    # pylint: disable=too-many-locals,too-many-branches
    def __handle_argparse_subparser_list(self, args):
        list_re = None
        if args.list_filter:
            list_re = re.compile(
                "^" + args.list_filter.replace("*", ".*").replace("?", ".") + "$"
            )

        show_rows = []
        ids = self.all_plugin_ids
        ids.sort()
        for next_plugin_id in ids:
            # if next_plugin_id.startswith("md9"):
            #     continue
            next_plugin_list = []
            for next_plugin in self.__registered_plugins:
                if next_plugin.plugin_id == next_plugin_id:
                    next_plugin_list.append(next_plugin)
            assert len(next_plugin_list) == 1
            next_plugin = next_plugin_list[0]
            if next_plugin.plugin_version == "0.0.0" and not args.show_all:
                continue
            does_match = True
            if list_re:
                does_match = list_re.match(next_plugin_id)
                if not does_match:
                    for next_name in next_plugin.plugin_names:
                        does_match = list_re.match(next_name)
                        if does_match:
                            break
            if does_match:
                is_enabled_now = next_plugin in self.__enabled_plugins
                display_row = [
                    next_plugin_id,
                    ", ".join(next_plugin.plugin_names),
                    str(next_plugin.plugin_enabled_by_default),
                    str(is_enabled_now),
                    next_plugin.plugin_version,
                ]
                show_rows.append(display_row)

        if show_rows:
            headers = [
                "id",
                "names",
                "enabled\n(default)",
                "enabled\n(current)",
                "version",
            ]
            table = columnar(show_rows, headers, no_borders=True)
            split_rows = table.split("\n")
            new_rows = []
            for next_row in split_rows:
                new_rows.append(next_row.rstrip())
            print("\n".join(new_rows))
        else:
            print(
                f"No plugin rule identifiers matches the pattern '{args.list_filter}'."
            )

    # pylint: enable=too-many-locals,too-many-branches

    def __handle_argparse_subparser_info(self, args):
        found_plugin = list(
            filter(
                lambda x: args.info_filter in x.plugin_identifiers,
                self.__registered_plugins,
            )
        )
        if not found_plugin:
            print(
                f"Unable to find a plugin with an id or name of '{args.info_filter}'."
            )
            return 1

        found_plugin = found_plugin[0]
        show_rows = []
        next_row = ["Id", found_plugin.plugin_id]
        show_rows.append(next_row)
        next_row = ["Name(s)", ",".join(found_plugin.plugin_names)]
        show_rows.append(next_row)
        next_row = ["Short Description", found_plugin.plugin_description]
        show_rows.append(next_row)
        if found_plugin.plugin_url:
            next_row = ["Description Url", found_plugin.plugin_url]
            show_rows.append(next_row)
        if found_plugin.plugin_configuration:
            next_row = ["Configuration Items", found_plugin.plugin_configuration]
            show_rows.append(next_row)

        headers = ["Item", "Description"]
        table = columnar(show_rows, headers, no_borders=True)
        split_rows = table.split("\n")
        new_rows = []
        for next_row in split_rows:
            new_rows.append(next_row.rstrip())
        print("\n".join(new_rows))
        return 0

    def handle_argparse_subparser(self, args):
        """
        Handle the parsing for this subparser.
        """
        subparser_value = getattr(args, PluginManager.__root_subparser_name)
        return_code = 0
        if subparser_value == "list":
            self.__handle_argparse_subparser_list(args)
        elif subparser_value == "info":
            return_code = self.__handle_argparse_subparser_info(args)
        else:
            PluginManager.__argparse_subparser.print_help()
            sys.exit(2)
        return return_code

    # pylint: disable=too-many-arguments
    def log_scan_failure(
        self,
        scan_file,
        line_number,
        column_number,
        rule_id,
        rule_name,
        rule_description,
        extra_error_information=None,
    ):
        """
        Log the scan failure in the appropriate format.
        """

        if self.__document_pragmas and line_number in self.__document_pragmas:
            id_set = self.__document_pragmas[line_number]
            rule_id = rule_id.lower()
            if rule_id in id_set:
                return

        extra_info = f" [{extra_error_information}]" if extra_error_information else ""

        print(
            "{0}:{1}:{2}: {3}: {4}{5} ({6})".format(
                scan_file,
                line_number,
                column_number,
                rule_id.upper(),
                rule_description,
                extra_info,
                rule_name,
            )
        )
        self.number_of_scan_failures += 1

    # pylint: enable=too-many-arguments

    def log_pragma_failure(self, scan_file, line_number, pragma_error):
        """
        Log the pragma failure in the appropriate format.
        """

        print("{0}:{1}:1: INLINE: {2}".format(scan_file, line_number, pragma_error))
        self.number_of_pragma_failures += 1

    # pylint: disable=too-many-locals
    def compile_pragmas(self, scan_file, pragma_lines):
        """
        Go through the list of extracted pragmas and compile them.
        """

        for next_line_number in pragma_lines.keys():
            PragmaExtension.compile_single_pragma(
                scan_file,
                next_line_number,
                pragma_lines,
                self.__all_ids,
                self.__document_pragmas,
                self.log_pragma_failure,
            )

    # pylint: enable=too-many-locals

    @classmethod
    def __find_eligible_plugins_in_directory(cls, directory_to_search):
        """
        Given a directory to search, scan for eligible modules to load later.
        """

        plugin_files = [
            x
            for x in os.listdir(directory_to_search)
            if x.endswith(".py") and x[0:-3] != "__init__"
        ]
        return plugin_files

    @classmethod
    def __snake_to_camel(cls, word):

        return "".join(x.capitalize() or "_" for x in word.split("_"))

    def __attempt_to_load_plugin(
        self, next_plugin_module, plugin_class_name, next_plugin_file
    ):
        """
        Attempt to cleanly load the specified plugin.
        """
        try:
            mod = __import__(next_plugin_module)
        except Exception as this_exception:
            raise BadPluginError(file_name=next_plugin_file) from this_exception

        if not hasattr(mod, plugin_class_name):
            raise BadPluginError(
                file_name=next_plugin_file, class_name=plugin_class_name
            ) from None
        my_class = getattr(mod, plugin_class_name)

        try:
            plugin_class_instance = my_class()
        except Exception as this_exception:
            raise BadPluginError(
                file_name=next_plugin_file,
                class_name=plugin_class_name,
                is_constructor=True,
            ) from this_exception
        self.__loaded_classes.append((plugin_class_instance, next_plugin_file))

    def __load_plugins(self, directory_to_search, plugin_files):
        """
        Given an array of discovered modules, load them into the global namespace.
        """

        if os.path.abspath(directory_to_search) not in sys.path:
            sys.path.insert(0, os.path.abspath(directory_to_search))

        for next_plugin_file in plugin_files:
            next_plugin_module = next_plugin_file[0:-3]
            plugin_class_name = self.__snake_to_camel(next_plugin_module)
            self.__attempt_to_load_plugin(
                next_plugin_module, plugin_class_name, next_plugin_file
            )

    def __determine_if_plugin_enabled(
        self,
        plugin_object,
        command_line_enabled_rules,
        command_line_disabled_rules,
        properties,
    ):
        """
        Given the enable and disable rule values, evaluate the enabled or disabled
        state of the plugin in proper order.
        """

        new_value = None
        LOGGER.debug(
            "Plugin '%s', identifiers: %s",
            plugin_object.plugin_id,
            plugin_object.plugin_identifiers,
        )

        if command_line_disabled_rules:
            LOGGER.debug(
                "Disabled on command line: %s", str(command_line_disabled_rules)
            )
            for next_identifier in plugin_object.plugin_identifiers:
                if next_identifier in command_line_disabled_rules:
                    new_value = False
                    LOGGER.debug("Plugin is disabled from command line.")
                    break
        if new_value is None and command_line_enabled_rules:
            LOGGER.debug("Enabled on command line: %s", str(command_line_enabled_rules))
            for next_identifier in plugin_object.plugin_identifiers:
                if next_identifier in command_line_enabled_rules:
                    new_value = True
                    LOGGER.debug("Plugin is enabled from command line.")
                    break
        if new_value is None:
            plugin_specific_facade = self.__find_configuration_for_plugin(
                plugin_object, properties
            )
            if plugin_specific_facade:
                LOGGER.debug(
                    "Plugins specific configuration found, searching for key 'enabled'."
                )
                new_value = plugin_specific_facade.get_boolean_property(
                    "enabled", default_value=None
                )
                if new_value:
                    LOGGER.debug(
                        "Plugin specific key 'enabled' found, value is '%s'.",
                        str(new_value),
                    )
        if new_value is None:
            LOGGER.debug(
                "No other enable state found, setting to default of '%s'.",
                str(plugin_object.plugin_enabled_by_default),
            )

        return (
            plugin_object.plugin_enabled_by_default if new_value is None else new_value
        )

    # pylint: enable=too-many-arguments

    @classmethod
    def __verify_string_field(cls, plugin_instance, field_name, field_value):
        """
        Verify that the detail field is a valid string.
        """

        if not isinstance(field_value, str):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )
        if not field_value:
            raise BadPluginError(
                class_name=type(plugin_instance).__name__,
                field_name=field_name,
                is_empty=True,
            )

    @classmethod
    def __verify_boolean_field(cls, plugin_instance, field_name, field_value):
        """
        Verify that the detail field is a valid boolean.
        """

        if not isinstance(field_value, bool):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )

    @classmethod
    def __verify_integer_field(cls, plugin_instance, field_name, field_value):
        """
        Verify that the detail field is a valid integer.
        """

        if not isinstance(field_value, int):
            raise BadPluginError(
                class_name=type(plugin_instance).__name__, field_name=field_name
            )

    def __get_plugin_details(self, plugin_instance, instance_file_name):
        """
        Query the plugin for details and verify that they are reasonable.
        """

        try:
            instance_details = plugin_instance.get_details()
            (
                plugin_id,
                plugin_name,
                plugin_description,
                plugin_enabled_by_default,
                plugin_version,
                plugin_interface_version,
                plugin_url,
                plugin_configuration,
            ) = (
                instance_details.plugin_id,
                instance_details.plugin_name,
                instance_details.plugin_description,
                instance_details.plugin_enabled_by_default,
                instance_details.plugin_version,
                instance_details.plugin_interface_version,
                instance_details.plugin_url,
                instance_details.plugin_configuration,
            )
        except Exception as this_exception:
            raise BadPluginError(
                class_name=type(plugin_instance).__name__,
            ) from this_exception

        self.__verify_string_field(plugin_instance, "plugin_id", plugin_id)
        self.__verify_string_field(plugin_instance, "plugin_name", plugin_name)
        self.__verify_string_field(
            plugin_instance, "plugin_description", plugin_description
        )
        self.__verify_boolean_field(
            plugin_instance, "plugin_enabled_by_default", plugin_enabled_by_default
        )
        self.__verify_string_field(plugin_instance, "plugin_version", plugin_version)
        self.__verify_integer_field(
            plugin_instance, "plugin_interface_version", plugin_interface_version
        )
        if plugin_interface_version != 1:
            raise BadPluginError(
                formatted_message=f"Plugin '{instance_file_name}' with an interface version ('{plugin_interface_version}') that is not '1'."
            )
        if plugin_url:
            self.__verify_string_field(plugin_instance, "plugin_url", plugin_url)
        if plugin_configuration:
            self.__verify_string_field(
                plugin_instance, "plugin_configuration", plugin_configuration
            )

        plugin_object = FoundPlugin(
            plugin_id,
            plugin_name,
            plugin_description,
            plugin_instance,
            plugin_enabled_by_default,
            plugin_version,
            plugin_interface_version,
            instance_file_name,
            plugin_url,
            plugin_configuration,
        )
        return plugin_object

    # pylint: disable=too-many-arguments
    def __register_individual_plugin(
        self,
        plugin_instance,
        instance_file_name,
        command_line_enabled_rules,
        command_line_disabled_rules,
        properties,
    ):
        """
        Register an individual plugin for use.
        """

        plugin_object = self.__get_plugin_details(plugin_instance, instance_file_name)

        next_key = plugin_object.plugin_id
        if not PluginManager.__id_regex.match(next_key):
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with id '{next_key}' as id is not a valid id in the form 'aannn' or 'aaannn'."
            )

        if next_key in self.__all_ids:
            found_plugin = self.__all_ids[next_key]
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with id '{next_key}' as plugin '{found_plugin.plugin_file_name}' is already registered with that id."
            )
        self.__all_ids[next_key] = plugin_object

        for next_key in plugin_object.plugin_names:
            if not PluginManager.__name_regex.match(next_key):
                raise ValueError(
                    f"Unable to register plugin '{instance_file_name}' with name '{next_key}' as name is not a valid name in the form 'an-an'."
                )
            if next_key in self.__all_ids:
                found_plugin = self.__all_ids[next_key]
                raise ValueError(
                    f"Unable to register plugin '{instance_file_name}' with name '{next_key}' as plugin '{found_plugin.plugin_file_name}' is already registered with that name."
                )
            self.__all_ids[next_key] = plugin_object
        if not plugin_object.plugin_description.strip():
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with a description string that is blank."
            )
        if not PluginManager.__version_regex.match(plugin_object.plugin_version):
            raise ValueError(
                f"Unable to register plugin '{instance_file_name}' with a version string that is not a valid semantic version."
            )

        self.__registered_plugins.append(plugin_object)
        if self.__determine_if_plugin_enabled(
            plugin_object,
            command_line_enabled_rules,
            command_line_disabled_rules,
            properties,
        ):
            self.__enabled_plugins.append(plugin_object)

    # pylint: enable=too-many-arguments

    def __register_plugins(
        self,
        enable_rules_from_command_line,
        disable_rules_from_command_line,
        properties,
    ):
        """
        Scan the global namespace for all subclasses of the 'Plugin' class to use as
        plugins.
        """

        (
            command_line_enabled_rules,
            command_line_disabled_rules,
            self.__registered_plugins,
            self.__enabled_plugins,
            self.__all_ids,
        ) = (set(), set(), [], [], {})
        if enable_rules_from_command_line:
            for next_rule_identifier in enable_rules_from_command_line.lower().split(
                ","
            ):
                command_line_enabled_rules.add(next_rule_identifier)
        if disable_rules_from_command_line:
            for next_rule_identifier in disable_rules_from_command_line.lower().split(
                ","
            ):
                command_line_disabled_rules.add(next_rule_identifier)

        for plugin_instance, instance_file_name in self.__loaded_classes:
            self.__register_individual_plugin(
                plugin_instance,
                instance_file_name,
                command_line_enabled_rules,
                command_line_disabled_rules,
                properties,
            )

    @property
    def all_plugin_ids(self):
        """
        Get a list of all plugins by their id.
        """

        id_list = []
        for next_plugin in self.__registered_plugins:
            id_list.append(next_plugin.plugin_id)
        return id_list

    @classmethod
    def __find_configuration_for_plugin(
        cls, next_plugin, properties, always_return_facade=False
    ):
        plugin_specific_facade = None
        first_facade = None
        for next_key_name in next_plugin.plugin_identifiers:
            plugin_section_title = f"{PluginManager.__plugin_prefix}{properties.separator}{next_key_name}{properties.separator}"
            section_facade_candidate = ApplicationPropertiesFacade(
                properties, plugin_section_title
            )
            if not first_facade:
                first_facade = section_facade_candidate
            if section_facade_candidate.property_names:
                plugin_specific_facade = section_facade_candidate
                break

        if always_return_facade and not plugin_specific_facade:
            plugin_specific_facade = first_facade
        return plugin_specific_facade

    def apply_configuration(self, properties):
        """
        Apply any supplied configuration to each of the enabled plugins.
        """

        (
            self.__enabled_plugins_for_starting_new_file,
            self.__enabled_plugins_for_next_token,
            self.__enabled_plugins_for_next_line,
            self.__enabled_plugins_for_completed_file,
        ) = ([], [], [], [])

        for next_plugin in self.__enabled_plugins:
            try:
                plugin_specific_facade = self.__find_configuration_for_plugin(
                    next_plugin, properties, always_return_facade=True
                )

                next_plugin.plugin_instance.set_configuration_map(
                    plugin_specific_facade
                )
                next_plugin.plugin_instance.initialize_from_config()
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

            if next_plugin.plugin_instance.is_next_token_implemented_in_plugin:
                self.__enabled_plugins_for_next_token.append(next_plugin)
            if next_plugin.plugin_instance.is_next_line_implemented_in_plugin:
                self.__enabled_plugins_for_next_line.append(next_plugin)
            if next_plugin.plugin_instance.is_completed_file_implemented_in_plugin:
                self.__enabled_plugins_for_completed_file.append(next_plugin)
            if next_plugin.plugin_instance.is_starting_new_file_implemented_in_plugin:
                self.__enabled_plugins_for_starting_new_file.append(next_plugin)

    def starting_new_file(self, file_being_started):
        """
        Inform any listeners that a new current file has been started.
        """
        self.__document_pragmas = {}
        for next_plugin in self.__enabled_plugins_for_starting_new_file:
            try:
                next_plugin.plugin_instance.starting_new_file()
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

        return ScanContext(self, file_being_started)

    def completed_file(self, context, line_number):
        """
        Inform any listeners that the current file has been completed.
        """
        context.line_number = line_number
        for next_plugin in self.__enabled_plugins_for_completed_file:
            try:
                next_plugin.plugin_instance.completed_file(context)
            except Exception as this_exception:
                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    cause=this_exception,
                ) from this_exception

    def next_line(self, context, line_number, line):
        """
        Inform any listeners that a new line has been loaded.
        """
        context.line_number = line_number
        for next_plugin in self.__enabled_plugins_for_next_line:
            try:
                next_plugin.plugin_instance.next_line(context, line)
            except Exception as this_exception:
                actual_line = line if self.__show_stack_trace else None

                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    line_number=line_number,
                    actual_line=actual_line,
                    cause=this_exception,
                ) from this_exception

    def next_token(self, context, token):
        """
        Inform any listeners of a new token that has been processed.
        """
        for next_plugin in self.__enabled_plugins_for_next_token:
            try:
                next_plugin.plugin_instance.next_token(context, token)
            except Exception as this_exception:
                actual_token = token if self.__show_stack_trace else None

                raise BadPluginError(
                    next_plugin.plugin_id,
                    inspect.stack()[0].function,
                    line_number=token.line_number,
                    column_number=token.column_number,
                    actual_token=actual_token,
                    cause=this_exception,
                ) from this_exception


# pylint: enable=too-many-instance-attributes



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_476():
    """
    Test case 476:  (part 1) Rule 14
    """

    # Arrange
    source_markdown = """***foo***"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[emphasis(1,2):2:*]",
        "[text(1,4):foo:]",
        "[end-emphasis(1,7)::]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em><strong>foo</strong></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_477():
    """
    Test case 477:  (part 2) Rule 14
    """

    # Arrange
    source_markdown = """_____foo_____"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[emphasis(1,2):2:_]",
        "[emphasis(1,4):2:_]",
        "[text(1,6):foo:]",
        "[end-emphasis(1,9)::]",
        "[end-emphasis(1,11)::]",
        "[end-emphasis(1,13)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em><strong><strong>foo</strong></strong></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_478():
    """
    Test case 478:  (part 1) Rule 15
    """

    # Arrange
    source_markdown = """*foo _bar* baz_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[text(1,6):_:]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[text(1,11): baz:]",
        "[text(1,15):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo _bar</em> baz_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_479():
    """
    Test case 479:  (part 2) Rule 15
    """

    # Arrange
    source_markdown = """*foo __bar *baz bim__ bam*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):2:_]",
        "[text(1,8):bar :]",
        "[text(1,12):*:]",
        "[text(1,13):baz bim:]",
        "[end-emphasis(1,20)::]",
        "[text(1,22): bam:]",
        "[end-emphasis(1,26)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <strong>bar *baz bim</strong> bam</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_480():
    """
    Test case 480:  (part 1) Rule 16
    """

    # Arrange
    source_markdown = """**foo **bar baz**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):**:]",
        "[text(1,3):foo :]",
        "[emphasis(1,7):2:*]",
        "[text(1,9):bar baz:]",
        "[end-emphasis(1,16)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>**foo <strong>bar baz</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_481():
    """
    Test case 481:  (part 2) Rule 16
    """

    # Arrange
    source_markdown = """*foo *bar baz*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):1:*]",
        "[text(1,7):bar baz:]",
        "[end-emphasis(1,14)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*foo <em>bar baz</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_482():
    """
    Test case 482:  (part 1) Rule 17
    """

    # Arrange
    source_markdown = """*[bar*](/url)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[link(1,2):inline:/url:::::bar*:False::::]",
        "[text(1,3):bar:]",
        "[text(1,6):*:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*<a href="/url">bar*</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_483():
    """
    Test case 483:  (part 2) Rule 17
    """

    # Arrange
    source_markdown = """_foo [bar_](/url)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2):foo :]",
        "[link(1,6):inline:/url:::::bar_:False::::]",
        "[text(1,7):bar:]",
        "[text(1,10):_:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_foo <a href="/url">bar_</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_484():
    """
    Test case 484:  (part 3) Rule 17
    """

    # Arrange
    source_markdown = """*<img src="foo" title="*"/>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        '[raw-html(1,2):img src="foo" title="*"/]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*<img src="foo" title="*"/></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_485():
    """
    Test case 485:  (part 4) Rule 17
    """

    # Arrange
    source_markdown = """**<a href="**">"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):**:]",
        '[raw-html(1,3):a href="**"]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>**<a href="**"></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_486():
    """
    Test case 486:  (part 5) Rule 17
    """

    # Arrange
    source_markdown = """__<a href="__">"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):__:]",
        '[raw-html(1,3):a href="__"]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>__<a href="__"></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_487():
    """
    Test case 487:  (part 6) Rule 17
    """

    # Arrange
    source_markdown = """*a `*`*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):a :]",
        "[icode-span(1,4):*:`::]",
        "[end-emphasis(1,7)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>a <code>*</code></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_488():
    """
    Test case 488:  (part 7) Rule 17
    """

    # Arrange
    source_markdown = """_a `_`_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):a :]",
        "[icode-span(1,4):_:`::]",
        "[end-emphasis(1,7)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>a <code>_</code></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_489():
    """
    Test case 489:  (part 8) Rule 17
    """

    # Arrange
    source_markdown = """**a<http://foo.bar/?q=**>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):**:]",
        "[text(1,3):a:]",
        "[uri-autolink(1,4):http://foo.bar/?q=**]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>**a<a href="http://foo.bar/?q=**">http://foo.bar/?q=**</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_490():
    """
    Test case 490:  (part 9) Rule 17
    """

    # Arrange
    source_markdown = """__a<http://foo.bar/?q=__>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):__:]",
        "[text(1,3):a:]",
        "[uri-autolink(1,4):http://foo.bar/?q=__]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>__a<a href="http://foo.bar/?q=__">http://foo.bar/?q=__</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#raw-html
"""

import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_raw_html_632():
    """
    Test case 632:  Here are some simple open tags:
    """

    # Arrange
    source_markdown = """<a><bab><c2c>"""
    expected_tokens = [
        "[para(1,1):]",
        "[raw-html(1,1):a]",
        "[raw-html(1,4):bab]",
        "[raw-html(1,9):c2c]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a><bab><c2c></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_632a():
    """
    Test case 632a:  Here are some simple open tags:
    """

    # Arrange
    source_markdown = """each <a> tag <bab> is <c2c> different"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):each :]",
        "[raw-html(1,6):a]",
        "[text(1,9): tag :]",
        "[raw-html(1,14):bab]",
        "[text(1,19): is :]",
        "[raw-html(1,23):c2c]",
        "[text(1,28): different:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>each <a> tag <bab> is <c2c> different</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_633():
    """
    Test case 633:  Empty elements:
    """

    # Arrange
    source_markdown = """<a/><b2/>"""
    expected_tokens = [
        "[para(1,1):]",
        "[raw-html(1,1):a/]",
        "[raw-html(1,5):b2/]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a/><b2/></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_634():
    """
    Test case 634:  Whitespace is allowed:
    """

    # Arrange
    source_markdown = """<a  /><b2
data="foo" >"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[raw-html(1,1):a  /]",
        '[raw-html(1,7):b2\ndata="foo" ]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a  /><b2
data="foo" ></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_634a():
    """
    Test case 634a:  variation
    """

    # Arrange
    source_markdown = """- <a  /><b2
data="foo" ><c>"""
    expected_tokens = [
        "[ulist(1,1):-::2::]",
        "[para(1,3):\n]",
        "[raw-html(1,3):a  /]",
        '[raw-html(1,9):b2\ndata="foo" ]',
        "[raw-html(2,13):c]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li><a  /><b2
data="foo" ><c></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_634bx():
    """
    Test case 634b:  variation
    """

    # Arrange
    source_markdown = """> <a  /><b2
data="foo" ><c>"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):\n]",
        "[raw-html(1,3):a  /]",
        '[raw-html(1,9):b2\ndata="foo" ]',
        "[raw-html(2,13):c]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a  /><b2
data="foo" ><c></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_634ba():
    """
    Test case 634ba:  variation
    """

    # Arrange
    source_markdown = """> <a  /><b2
> data="foo" ><c>"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):\n]",
        "[raw-html(1,3):a  /]",
        '[raw-html(1,9):b2\ndata="foo" ]',
        "[raw-html(2,15):c]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a  /><b2
data="foo" ><c></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, disable_consistency_checks=True
    )


@pytest.mark.gfm
def test_raw_html_634bb():
    """
    Test case 634bb:  variation
    """

    # Arrange
    source_markdown = """> <a  /><b2
>   data="foo" ><c>"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> ]",
        "[para(1,3):\n  ]",
        "[raw-html(1,3):a  /]",
        '[raw-html(1,9):b2\n\a  \a\x03\adata="foo" ]',
        "[raw-html(2,17):c]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><a  /><b2
data="foo" ><c></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(
        source_markdown, expected_gfm, expected_tokens, disable_consistency_checks=True
    )


@pytest.mark.gfm
def test_raw_html_634c():
    """
    Test case 634c:  variation
    """

    # Arrange
    source_markdown = """<a  /><b2
data="foo" ><c>
---"""
    expected_tokens = [
        "[setext(3,1):-:3::(1,1)]",
        "[raw-html(1,1):a  /]",
        '[raw-html(1,7):b2\ndata="foo" ]',
        "[raw-html(2,13):c]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2><a  /><b2
data="foo" ><c></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_634d():
    """
    Test case 634c:  variation
    """

    # Arrange
    source_markdown = """# <a  /><b2
data="foo" ><c>"""
    expected_tokens = [
        "[atx(1,1):1:0:]",
        "[text(1,3)::\a \a\x03\a]",
        "[raw-html(1,3):a  /]",
        "[text(1,9):\a<\a&lt;\ab2:]",
        "[end-atx::]",
        "[para(2,1):]",
        '[text(2,1):data=\a"\a&quot;\afoo\a"\a&quot;\a \a>\a&gt;\a:]',
        "[raw-html(2,13):c]",
        "[end-para:::True]",
    ]
    expected_gfm = """<h1><a  />&lt;b2</h1>
<p>data=&quot;foo&quot; &gt;<c></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_635():
    """
    Test case 635:  With attributes:
    """

    # Arrange
    source_markdown = """<a foo="bar" bam = 'baz <em>"</em>'
_boolean zoop:33=zoop:33 />"""
    expected_tokens = [
        "[para(1,1):\n]",
        """[raw-html(1,1):a foo="bar" bam = 'baz <em>"</em>'
_boolean zoop:33=zoop:33 /]""",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a foo="bar" bam = 'baz <em>"</em>'
_boolean zoop:33=zoop:33 /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_636():
    """
    Test case 636:  Custom tag names can be used:
    """

    # Arrange
    source_markdown = """Foo <responsive-image src="foo.jpg" />"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):Foo :]",
        '[raw-html(1,5):responsive-image src="foo.jpg" /]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo <responsive-image src="foo.jpg" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_637():
    """
    Test case 637:  Illegal tag names, not parsed as HTML:
    """

    # Arrange
    source_markdown = """<33> <__>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\a<\a&lt;\a33\a>\a&gt;\a \a<\a&lt;\a:]",
        "[text(1,7):__:]",
        "[text(1,9):\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;33&gt; &lt;__&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_638():
    """
    Test case 638:  Illegal attribute names:
    """

    # Arrange
    source_markdown = """<a h*#ref="hi">"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\a<\a&lt;\aa h:]",
        "[text(1,5):*:]",
        '[text(1,6):#ref=\a"\a&quot;\ahi\a"\a&quot;\a\a>\a&gt;\a:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;a h*#ref=&quot;hi&quot;&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_639():
    """
    Test case 639:  Illegal attribute values:
    """

    # Arrange
    source_markdown = """<a href="hi'> <a href=hi'>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\a<\a&lt;\aa href=\a\"\a&quot;\ahi'\a>\a&gt;\a \a<\a&lt;\aa href=hi'\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;a href=&quot;hi'&gt; &lt;a href=hi'&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_639a():
    """
    Test case 639a:  Illegal attribute values:
    """

    # Arrange
    source_markdown = """<a href='hi"> <a href=hi">"""
    expected_tokens = [
        "[para(1,1):]",
        '[text(1,1):\a<\a&lt;\aa href=\'hi\a"\a&quot;\a\a>\a&gt;\a \a<\a&lt;\aa href=hi\a"\a&quot;\a\a>\a&gt;\a:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;a href='hi&quot;&gt; &lt;a href=hi&quot;&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_640():
    """
    Test case 640:  Illegal whitespace:
    """

    # Arrange
    source_markdown = """< a><
foo><bar/ >
<foo bar=baz
bim!bop />"""
    expected_tokens = [
        "[para(1,1):\n\n\n]",
        "[text(1,1):\a<\a&lt;\a a\a>\a&gt;\a\a<\a&lt;\a\nfoo\a>\a&gt;\a\a<\a&lt;\abar/ \a>\a&gt;\a\n\a<\a&lt;\afoo bar=baz\nbim!bop /\a>\a&gt;\a::\n\n\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt; a&gt;&lt;
foo&gt;&lt;bar/ &gt;
&lt;foo bar=baz
bim!bop /&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_641():
    """
    Test case 641:  Missing whitespace:
    """

    # Arrange
    source_markdown = """<a href='bar'title=title>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\a<\a&lt;\aa href='bar'title=title\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;a href='bar'title=title&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_642():
    """
    Test case 642:  Closing tags:
    """

    # Arrange
    source_markdown = """</a></foo >"""
    expected_tokens = [
        "[para(1,1):]",
        "[raw-html(1,1):/a]",
        "[raw-html(1,5):/foo ]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p></a></foo ></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_642a():
    """
    Test case 642a:  closing tag character without a valid closing tag name
    """

    # Arrange
    source_markdown = """</>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):\a<\a&lt;\a/\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;/&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_643():
    """
    Test case 643:  Illegal attributes in closing tag:
    """

    # Arrange
    source_markdown = """</a href="foo">"""
    expected_tokens = [
        "[para(1,1):]",
        '[text(1,1):\a<\a&lt;\a/a href=\a"\a&quot;\afoo\a"\a&quot;\a\a>\a&gt;\a:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;/a href=&quot;foo&quot;&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_644():
    """
    Test case 644:  (part 1) Comments:
    """

    # Arrange
    source_markdown = """foo <!-- this is a
comment - with hyphen -->"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):!-- this is a\ncomment - with hyphen --]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <!-- this is a
comment - with hyphen --></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_644a():
    """
    Test case 644a:  variation with space
    """

    # Arrange
    source_markdown = """foo <!-- this is a
 comment - with hyphen -->"""
    expected_tokens = [
        "[para(1,1):\n ]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):!-- this is a\n\a \a\x03\acomment - with hyphen --]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <!-- this is a
comment - with hyphen --></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_644aa():
    """
    Test case 644aa:  variation with space
    """

    # Arrange
    source_markdown = """foo <!-- this is a
 comment - with hyphen -->bar"""
    expected_tokens = [
        "[para(1,1):\n ]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):!-- this is a\n\a \a\x03\acomment - with hyphen --]",
        "[text(2,27):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <!-- this is a
comment - with hyphen -->bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_644b():
    """
    Test case 644b:  variation with space
    """

    # Arrange
    source_markdown = """foo <!-- this is a
 comment - with hyphen -->  foo
bar"""
    expected_tokens = [
        "[para(1,1):\n \n]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):!-- this is a\n\a \a\x03\acomment - with hyphen --]",
        "[text(2,27):  foo\nbar::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <!-- this is a
comment - with hyphen -->  foo
bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_644c():
    """
    Test case 644c:  variation with space
    """

    # Arrange
    source_markdown = """foo <!--
this is a
comment - with hyphen
-->bar"""
    expected_tokens = [
        "[para(1,1):\n\n\n]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):!--\nthis is a\ncomment - with hyphen\n--]",
        "[text(4,4):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <!--
this is a
comment - with hyphen
-->bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_645():
    """
    Test case 645:  (part 2) Comments:
    """

    # Arrange
    source_markdown = """foo <!-- not a comment -- two hyphens -->"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo \a<\a&lt;\a!-- not a comment -- two hyphens --\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo &lt;!-- not a comment -- two hyphens --&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_646():
    """
    Test case 646:  Not comments:
    """

    # Arrange
    source_markdown = """foo <!--> foo -->

foo <!-- foo--->"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo \a<\a&lt;\a!--\a>\a&gt;\a foo --\a>\a&gt;\a:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):foo \a<\a&lt;\a!-- foo---\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo &lt;!--&gt; foo --&gt;</p>
<p>foo &lt;!-- foo---&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_646a():
    """
    Test case 646a:  Not comments:
    """

    # Arrange
    source_markdown = """foo <![CDATA[>&<>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo \a<\a&lt;\a:]",
        "[text(1,6):![:]",
        "[text(1,8):CDATA:]",
        "[text(1,13):[:]",
        "[text(1,14):\a>\a&gt;\a\a&\a&amp;\a\a<\a&lt;\a\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo &lt;![CDATA[&gt;&amp;&lt;&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_647():
    """
    Test case 647:  Processing instructions:
    """

    # Arrange
    source_markdown = """foo <?php echo $a; ?>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):?php echo $a; ?]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <?php echo $a; ?></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_648():
    """
    Test case 648:  Declarations:
    """

    # Arrange
    source_markdown = """foo <!ELEMENT br EMPTY>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):!ELEMENT br EMPTY]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <!ELEMENT br EMPTY></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_648a():
    """
    Test case 648:  Declarations:
    """

    # Arrange
    source_markdown = """foo <!ELEMENT>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo \a<\a&lt;\a!ELEMENT\a>\a&gt;\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo &lt;!ELEMENT&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_649():
    """
    Test case 649:  CDATA sections:
    """

    # Arrange
    source_markdown = """foo <![CDATA[>&<]]>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[raw-html(1,5):![CDATA[>&<]]]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <![CDATA[>&<]]></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_650():
    """
    Test case 650:  Entity and numeric character references are preserved in HTML attributes:
    """

    # Arrange
    source_markdown = """foo <a href="&ouml;">"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        '[raw-html(1,5):a href="&ouml;"]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <a href="&ouml;"></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_651():
    """
    Test case 651:  (part 1) Backslash escapes do not work in HTML attributes:
    """

    # Arrange
    source_markdown = """foo <a href="\\*">"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        '[raw-html(1,5):a href="\\*"]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <a href="\\*"></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_raw_html_652():
    """
    Test case 652:  (part 2) Backslash escapes do not work in HTML attributes:
    """

    # Arrange
    source_markdown = """<a href="\\"">"""
    expected_tokens = [
        "[para(1,1):]",
        '[text(1,1):\a<\a&lt;\aa href=\a"\a&quot;\a\\\b\a"\a&quot;\a\a"\a&quot;\a\a>\a&gt;\a:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>&lt;a href=&quot;&quot;&quot;&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to implement a plugin that ensures that the indentation for List Items
are equivalent with each other.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd005(Plugin):
    """
    Class to implement a plugin that ensures that the indentation for List Items
    are equivalent with each other.
    """

    def __init__(self):
        super().__init__()
        self.__list_stack = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="list-indent",
            plugin_id="MD005",
            plugin_enabled_by_default=True,
            plugin_description="Inconsistent indentation for list items at the same level",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md005.md",
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__list_stack = []

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_unordered_list_start or token.is_ordered_list_start:
            self.__list_stack.append(token)
        elif token.is_unordered_list_end or token.is_ordered_list_end:
            del self.__list_stack[-1]
        elif token.is_new_list_item:
            if self.__list_stack[-1].is_unordered_list_start:
                if self.__list_stack[-1].indent_level != token.indent_level:
                    extra_data = f"Expected: {len(self.__list_stack[-1].extracted_whitespace)}; Actual: {len(token.extracted_whitespace)}"
                    self.report_next_token_error(context, token, extra_data)
            elif self.__list_stack[-1].column_number != token.column_number:
                original_text = self.__list_stack[-1].list_start_content
                if self.__list_stack[-1].extracted_whitespace:
                    original_text += self.__list_stack[-1].extracted_whitespace
                original_text_length = len(original_text)
                current_prefix_length = len(
                    token.list_start_content + token.extracted_whitespace
                )
                if original_text_length == current_prefix_length:
                    assert token.indent_level == self.__list_stack[-1].indent_level
                else:
                    extra_data = f"Expected: {len(self.__list_stack[-1].extracted_whitespace)}; Actual: {len(token.extracted_whitespace)}"
                    self.report_next_token_error(context, token, extra_data)



"""
Module to implement a plugin that looks for code blocks for scripts that do not
look right.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd014(Plugin):
    """
    Class to implement a plugin that looks for code blocks for scripts that do not
    look right.
    """

    def __init__(self):
        super().__init__()
        self.__in_code_block = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="commands-show-output",
            plugin_id="MD014",
            plugin_enabled_by_default=True,
            plugin_description="Dollar signs used before commands without showing output",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md014.md",
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__in_code_block = False

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_code_block:
            self.__in_code_block = True
        elif token.is_code_block_end:
            self.__in_code_block = False
        elif self.__in_code_block and token.is_text:
            are_all_preceded_with_dollar_sign = True
            split_token_text = token.token_text.split("\n")
            for next_line in split_token_text:
                if not next_line.strip().startswith("$"):
                    are_all_preceded_with_dollar_sign = False
                    break
            if are_all_preceded_with_dollar_sign:
                self.report_next_token_error(context, token)



"""
Module to provide tests related to the MD040 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest


@pytest.mark.rules
def test_md040_good_fenced_block_with_language():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md040/good_fenced_block_with_language.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md040_good_fenced_block_with_space_before_language():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md040/good_fenced_block_with_space_before_language.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md040_bad_fenced_block_with_no_language():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md040/bad_fenced_block_with_no_language.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md040/bad_fenced_block_with_no_language.md:3:1: "
        + "MD040: Fenced code blocks should have a language specified (fenced-code-language)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md040_bad_fenced_block_with_whitespace():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md040/bad_fenced_block_with_whitespace.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md040/bad_fenced_block_with_whitespace.md:4:1: "
        + "MD040: Fenced code blocks should have a language specified (fenced-code-language)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_413():
    """
    Test case 413:  (part 1) Any nonempty sequence of inline elements can be the contents of an emphasized span.
    """

    # Arrange
    source_markdown = """*foo [bar](/url)*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[link(1,6):inline:/url:::::bar:False::::]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[end-emphasis(1,17)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <a href="/url">bar</a></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_414():
    """
    Test case 414:  (part 2) Any nonempty sequence of inline elements can be the contents of an emphasized span.
    """

    # Arrange
    source_markdown = """*foo
bar*"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo\nbar::\n]",
        "[end-emphasis(2,4)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo
bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_415():
    """
    Test case 415:  (part 1) In particular, emphasis and strong emphasis can be nested inside emphasis:
    """

    # Arrange
    source_markdown = """_foo __bar__ baz_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):2:_]",
        "[text(1,8):bar:]",
        "[end-emphasis(1,11)::]",
        "[text(1,13): baz:]",
        "[end-emphasis(1,17)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <strong>bar</strong> baz</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_416():
    """
    Test case 416:  (part 2) In particular, emphasis and strong emphasis can be nested inside emphasis:
    """

    # Arrange
    source_markdown = """_foo _bar_ baz_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):1:_]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[text(1,11): baz:]",
        "[end-emphasis(1,15)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <em>bar</em> baz</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_417():
    """
    Test case 417:  (part 3) In particular, emphasis and strong emphasis can be nested inside emphasis:
    """

    # Arrange
    source_markdown = """__foo_ bar_"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[emphasis(1,2):1:_]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[text(1,7): bar:]",
        "[end-emphasis(1,11)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em><em>foo</em> bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_418():
    """
    Test case 418:  (part 4) In particular, emphasis and strong emphasis can be nested inside emphasis:
    """

    # Arrange
    source_markdown = """*foo *bar**"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):1:*]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[end-emphasis(1,11)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <em>bar</em></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_419():
    """
    Test case 419:  (part 5) In particular, emphasis and strong emphasis can be nested inside emphasis:
    """

    # Arrange
    source_markdown = """*foo **bar** baz*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):2:*]",
        "[text(1,8):bar:]",
        "[end-emphasis(1,11)::]",
        "[text(1,13): baz:]",
        "[end-emphasis(1,17)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <strong>bar</strong> baz</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_420():
    """
    Test case 420:  (part 6) In particular, emphasis and strong emphasis can be nested inside emphasis:
    """

    # Arrange
    source_markdown = """*foo**bar**baz*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo:]",
        "[emphasis(1,5):2:*]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[text(1,12):baz:]",
        "[end-emphasis(1,15)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo<strong>bar</strong>baz</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_421():
    """
    Test case 421:  For the same reason, we don’t get two consecutive emphasis sections in this example:
    """

    # Arrange
    source_markdown = """*foo**bar*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo:]",
        "[text(1,5):**:]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo**bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_422():
    """
    Test case 422:  (part 1) The same condition ensures that the following cases are all strong emphasis nested inside emphasis, even when the interior spaces are omitted:
    """

    # Arrange
    source_markdown = """***foo** bar*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[emphasis(1,2):2:*]",
        "[text(1,4):foo:]",
        "[end-emphasis(1,7)::]",
        "[text(1,9): bar:]",
        "[end-emphasis(1,13)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em><strong>foo</strong> bar</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_423():
    """
    Test case 423:  (part 2) The same condition ensures that the following cases are all strong emphasis nested inside emphasis, even when the interior spaces are omitted:
    """

    # Arrange
    source_markdown = """*foo **bar***"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):2:*]",
        "[text(1,8):bar:]",
        "[end-emphasis(1,11)::]",
        "[end-emphasis(1,13)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <strong>bar</strong></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_424():
    """
    Test case 424:  (part 3) The same condition ensures that the following cases are all strong emphasis nested inside emphasis, even when the interior spaces are omitted:
    """

    # Arrange
    source_markdown = """*foo**bar***"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo:]",
        "[emphasis(1,5):2:*]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[end-emphasis(1,12)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo<strong>bar</strong></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_425():
    """
    Test case 425:  (part 1) When the lengths of the interior closing and opening delimiter runs are both multiples of 3, though, they can match to create emphasis:
    """

    # Arrange
    source_markdown = """foo***bar***baz"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo:]",
        "[emphasis(1,4):1:*]",
        "[emphasis(1,5):2:*]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[end-emphasis(1,12)::]",
        "[text(1,13):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo<em><strong>bar</strong></em>baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_426():
    """
    Test case 426:  (part 2) When the lengths of the interior closing and opening delimiter runs are both multiples of 3, though, they can match to create emphasis:
    """

    # Arrange
    source_markdown = """foo******bar*********baz"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo:]",
        "[emphasis(1,4):2:*]",
        "[emphasis(1,6):2:*]",
        "[emphasis(1,8):2:*]",
        "[text(1,10):bar:]",
        "[end-emphasis(1,13)::]",
        "[end-emphasis(1,15)::]",
        "[end-emphasis(1,17)::]",
        "[text(1,19):***:]",
        "[text(1,22):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>foo<strong><strong><strong>bar</strong></strong></strong>***baz</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_427():
    """
    Test case 427:  (part 1) When the lengths of the interior closing and opening delimiter runs are both multiples of 3, though, they can match to create emphasis:
    """

    # Arrange
    source_markdown = """*foo **bar *baz* bim** bop*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):2:*]",
        "[text(1,8):bar :]",
        "[emphasis(1,12):1:*]",
        "[text(1,13):baz:]",
        "[end-emphasis(1,16)::]",
        "[text(1,17): bim:]",
        "[end-emphasis(1,21)::]",
        "[text(1,23): bop:]",
        "[end-emphasis(1,27)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <strong>bar <em>baz</em> bim</strong> bop</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_428():
    """
    Test case 428:  (part 2) When the lengths of the interior closing and opening delimiter runs are both multiples of 3, though, they can match to create emphasis:
    """

    # Arrange
    source_markdown = """*foo [*bar*](/url)*"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[link(1,6):inline:/url:::::*bar*:False::::]",
        "[emphasis(1,7):1:*]",
        "[text(1,8):bar:]",
        "[end-emphasis(1,11)::]",
        "[end-link::]",
        "[end-emphasis(1,19)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo <a href="/url"><em>bar</em></a></em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_429():
    """
    Test case 429:  (part 1) There can be no empty emphasis or strong emphasis:
    """

    # Arrange
    source_markdown = """** is not an empty emphasis"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):**:]",
        "[text(1,3): is not an empty emphasis:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>** is not an empty emphasis</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_430():
    """
    Test case 430:  (part 2) There can be no empty emphasis or strong emphasis:
    """

    # Arrange
    source_markdown = """**** is not an empty strong emphasis"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):****:]",
        "[text(1,5): is not an empty strong emphasis:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>**** is not an empty strong emphasis</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Tests for the calculate_length function
"""
from pymarkdown.parser_helper import ParserHelper


def test_calculate_length_empty_string():
    """
    Make sure that an empty string is handled properly.
    """

    # Arrange
    input_string = ""
    expected_output = 0

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_single_space():
    """
    Make sure that a string with a single space is handled properly.
    """

    # Arrange
    input_string = " "
    expected_output = 1

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_multiple_spaces():
    """
    Make sure that a string with multiple spaces is handled properly.
    """

    # Arrange
    for i in range(1, 10):
        input_string = "".rjust(i, " ")
        expected_output = i

        # Act
        actual_output = ParserHelper.calculate_length(input_string)

        # Assert
        assert expected_output == actual_output


def test_calculate_length_single_tab():
    """
    Make sure that a string with a single tab is handled properly.
    """

    # Arrange
    input_string = "\t"
    expected_output = 4

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_multiple_tabs():
    """
    Make sure that a string with multiple tabs is handled properly.
    """

    # Arrange
    for i in range(1, 10):
        input_string = "".rjust(i, "\t")
        expected_output = i * 4

        # Act
        actual_output = ParserHelper.calculate_length(input_string)

        # Assert
        assert expected_output == actual_output


def test_calculate_length_space_then_tab():
    """
    Make sure that a string with a single space then a tab is handled properly.
    """

    # Arrange
    input_string = " \t"
    expected_output = 4

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_space_then_tab_twice():
    """
    Make sure that a string with a single space then a tab (twice) is handled properly.
    """

    # Arrange
    input_string = " \t"
    input_string += input_string
    expected_output = 8

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_double_space_then_tab():
    """
    Make sure that a string with two spaces then a tab is handled properly.
    """

    # Arrange
    input_string = "  \t"
    expected_output = 4

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_triple_space_then_tab():
    """
    Make sure that a string with three spaces then a tab is handled properly.
    """

    # Arrange
    input_string = "   \t"
    expected_output = 4

    # Act
    actual_output = ParserHelper.calculate_length(input_string)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_tab_after_0_index_start():
    """
    Make sure that a string with a tab is handled properly after a start of 0.
    """

    # Arrange
    input_string = "\t"
    start_index = 0
    expected_output = 4

    # Act
    actual_output = ParserHelper.calculate_length(input_string, start_index)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_tab_after_1_index_start():
    """
    Make sure that a string with a tab is handled properly after a start of 1.
    Note that with a start of 1, a tab moves it to the next tab stop at 4, specifying
    that 3 space characters should be added.
    """

    # Arrange
    input_string = "\t"
    start_index = 1
    expected_output = 3

    # Act
    actual_output = ParserHelper.calculate_length(input_string, start_index)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_tab_after_2_index_start():
    """
    Make sure that a string with a tab is handled properly after a start of 2.
    Note that with a start of 2, a tab moves it to the next tab stop at 4, specifying
    that 2 space characters should be added.
    """

    # Arrange
    input_string = "\t"
    start_index = 2
    expected_output = 2

    # Act
    actual_output = ParserHelper.calculate_length(input_string, start_index)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_tab_after_3_index_start():
    """
    Make sure that a string with a tab is handled properly after a start of 3.
    Note that with a start of 3, a tab moves it to the next tab stop at 4, specifying
    that 1 space characters should be added.
    """

    # Arrange
    input_string = "\t"
    start_index = 3
    expected_output = 1

    # Act
    actual_output = ParserHelper.calculate_length(input_string, start_index)

    # Assert
    assert expected_output == actual_output


def test_calculate_length_tab_after_4_index_start():
    """
    Make sure that a string with a tab is handled properly after a start of 4.
    Note that with a start of 4, a tab moves it to the next tab stop at 8, specifying
    that 4 space characters should be added.
    """

    # Arrange
    input_string = "\t"
    start_index = 4
    expected_output = 4

    # Act
    actual_output = ParserHelper.calculate_length(input_string, start_index)

    # Assert
    assert expected_output == actual_output



"""
Module to implement a plugin that looks for multiple blank lines in the files.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd012(Plugin):
    """
    Class to implement a plugin that looks for multiple blank lines in the files.
    """

    def __init__(self):
        super().__init__()
        self.__blank_lines_maximum = None
        self.__blank_line_count = None
        self.__last_blank_line = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-multiple-blanks",
            plugin_id="MD012",
            plugin_enabled_by_default=True,
            plugin_description="Multiple consecutive blank lines",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md012.md",
            plugin_configuration="maximum",
        )

    @classmethod
    def __validate_maximum(cls, found_value):
        if found_value < 0:
            raise ValueError("Allowable values are any non-negative integers.")

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """
        self.__blank_lines_maximum = self.plugin_configuration.get_integer_property(
            "maximum",
            default_value=1,
            valid_value_fn=self.__validate_maximum,
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__blank_line_count = 0
        self.__last_blank_line = None

    def __check_for_excess_blank_lines(self, context):
        if self.__blank_line_count > self.__blank_lines_maximum:
            extra_data = f"Expected: {self.__blank_lines_maximum}, Actual: {self.__blank_line_count}"

            self.report_next_token_error(
                context, self.__last_blank_line, extra_error_information=extra_data
            )

    def completed_file(self, context):
        """
        Event that the file being currently scanned is now completed.
        """
        self.__check_for_excess_blank_lines(context)

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_blank_line:
            self.__last_blank_line = token
            self.__blank_line_count += 1
        else:
            self.__check_for_excess_blank_lines(context)
            self.__blank_line_count = 0
